#include <string.h>
#include <math.h>
#include <pthread.h>

#include "rootvis.h"

extern Window ToonGetRootWindow(Display*, int, Window*);
extern void config_read(void);
extern void config_write(void);

// Forward declarations
static void rootvis_init(void);
static void rootvis_cleanup(void);
static void rootvis_about(void);
static void rootvis_configure(void);
static void rootvis_playback_start(void);
static void rootvis_playback_stop(void);
static void rootvis_render_freq(gint16 freq_data[2][256]);

// Callback functions
VisPlugin rootvis_vtable = {
	0, // Handle, filled in by xmms
	0, // Filename, filled in by xmms

	0,                     // Session ID
	"Root Spectrum Analyzer 0.0.5",  // description

	0, // # of PCM channels for render_pcm()
	1, // # of freq channels wanted for render_freq()    == MONO?

	rootvis_init,           // Called when plugin is enabled
	rootvis_cleanup,        // Called when plugin is disabled
	NULL,//rootvis_about,          // Show the about box
	rootvis_configure,      // Show the configure box
	0,                     // Called to disable plugin, filled in by xmms
	rootvis_playback_start, // Called when playback starts
	rootvis_playback_stop,  // Called when playback stops
	0,                     // Render the PCM data, must return quickly
	rootvis_render_freq     // Render the freq data, must return quickly
};

// XMMS entry point
VisPlugin *get_vplugin_info(void) {
	return &rootvis_vtable;
}

// X related
struct rootvis_x {
	Display *display;
	int screen;
	Window rootWin, Parent;
	GC gc;
} draw;

// thread talk

pthread_t work_thread;
pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;
enum {GO, CONF, STOP} thread_control;

gint16 shared_freq_data[2][256];

// Some helper stuff

void clean_data(void) {
	pthread_mutex_lock(&mutex1);
	memset(shared_freq_data, 0, sizeof(gint16) * 2 * 256);
	pthread_mutex_unlock(&mutex1);
}

void print_status(char msg[]) {
	if (conf.debug == 1) printf(">> rootvis >> %s\n", msg); // for debug purposes, but doesn't tell much anyway
}

void error_exit(char msg[]) {
	printf("*** ERROR (rootvis): %s\n", msg);
	rootvis_vtable.disable_plugin(&rootvis_vtable);
}

void initialize_X(void) {
	// XGCValues values;
	// unsigned long valuemask;
	print_status("Opening X Display");
	draw.display = XOpenDisplay(NULL);
	if (draw.display == NULL) {
		fprintf(stderr, "cannot connect to X server %s\n",
			getenv("DISPLAY") ? getenv("DISPLAY") : "(default)");
		error_exit("Connecting to X server failed");
	}
	print_status("Getting screen and window");
	draw.screen = DefaultScreen(draw.display);
	draw.rootWin = ToonGetRootWindow(draw.display, draw.screen, &draw.Parent);

	/* these variables are used to specify various attributes for the GC. */
	/* initial values for the GC. */
	//values.cap_style = CapButt;
	//values.join_style = JoinBevel;
	/* which values in 'values' to check when creating the GC. */
	// valuemask = NULL;
	print_status("Creating Graphical Context");
	draw.gc = XCreateGC(draw.display, draw.rootWin, 0, NULL);
	if (draw.gc < 0) {
		error_exit("Creating Graphical Context failed");
	}
	print_status("Setting Line Attributes");
	XSetLineAttributes(draw.display, draw.gc, 1, LineSolid, CapButt, JoinBevel);
}

long get_color(char col[3]) {
	Colormap def_colormap;
 	XColor dacolor;
	dacolor.red = col[RED] * 256;
	dacolor.green = col[GREEN] * 256;
	dacolor.blue = col[BLUE] * 256;
	print_status("Getting Colormap");
	def_colormap = DefaultColormap(draw.display, DefaultScreen(draw.display));
	print_status("Allocationg color");
	if (XAllocColor(draw.display, def_colormap, &dacolor) == 0) {
		error_exit("Color allocation failed");
	}
	return dacolor.pixel;
}

void draw_bar(int i, int color, int peakcolor, int shcolor, unsigned short level, unsigned short oldlevel, unsigned short peak, unsigned short oldpeak) {
	
	/* to make following cleaner, we work with redundant helper variables
	   this also avoids some calculations */
	register int a, b, c, d;
	int drawpeak = 0;
	
	if ((conf.peak.enabled)&&(peak != oldpeak)) drawpeak = 1;
	
	if (conf.geo.orientation < 2) { a = conf.geo.posx + i*(conf.bar.width+conf.bar.shadow+conf.geo.space); c = conf.bar.width; }
	else	{ b = conf.geo.posy + (conf.cutoff/conf.div - i)*(conf.bar.width+conf.bar.shadow+conf.geo.space); d = conf.bar.width; }
	
	if (level > oldlevel) {
		if ((conf.peak.enabled)&&(peak == level)) level--;
		if (conf.geo.orientation == 0) {	b = conf.geo.posy + conf.geo.height - level; d = level; }
		else if (conf.geo.orientation == 1) { b = conf.geo.posy; d = level; }
		else if (conf.geo.orientation == 2) { a = conf.geo.posx; c = level; }
		else	{ a = conf.geo.posx + conf.geo.height - level; c = level; }
		XSetForeground(draw.display, draw.gc, shcolor);
		XFillRectangle(draw.display, draw.rootWin, draw.gc, a+conf.bar.shadow, b+conf.bar.shadow, c, d);
		XSetForeground(draw.display, draw.gc, color);
		XFillRectangle(draw.display, draw.rootWin, draw.gc, a, b, c, d);
		if ((conf.peak.enabled)&&(peak == level)) level++;
	}
	if ((drawpeak == 1)&&(level >= oldlevel)&&(level < oldpeak)) {
		if (conf.geo.orientation == 0) {	b = conf.geo.posy + conf.geo.height - oldpeak; d = 1; }
		else if (conf.geo.orientation == 1) { b = conf.geo.posy + oldpeak; d = 1; }
		else if (conf.geo.orientation == 2) { a = conf.geo.posx + oldpeak; c = 1; }
		else	{ a = conf.geo.posx + conf.geo.height - oldpeak; c = 1; }
		XClearArea(draw.display, draw.rootWin, a, b, c, d, False);
	}
	if (level < oldlevel) {
		/* do me the pleasure and erase both at once */
		if (conf.peak.enabled) {
			if (oldlevel < oldpeak) oldlevel = oldpeak;
			if (oldlevel == peak) oldlevel--;
		}
				
		if (conf.geo.orientation == 0) { b = conf.geo.posy + conf.geo.height - oldlevel; d = oldlevel - level; }
		else if (conf.geo.orientation == 1) { b = conf.geo.posy + level; d = oldlevel - level; }
		else if (conf.geo.orientation == 2) { a = conf.geo.posx + level; c = oldlevel - level; }
		else	{ a = conf.geo.posx + conf.geo.height - oldlevel; c = oldlevel - level; }
		XClearArea(draw.display, draw.rootWin, a, b, c+conf.bar.shadow, d+conf.bar.shadow, False);
	}
	if ((drawpeak == 1)&&(peak > 0)) {
		XSetForeground(draw.display, draw.gc, peakcolor);
		if (conf.geo.orientation == 0) {	b = conf.geo.posy + conf.geo.height - peak; d = 1; }
		else if (conf.geo.orientation == 1) { b = conf.geo.posy + peak; d = 1; }
		else if (conf.geo.orientation == 2) { a = conf.geo.posx + peak; c = 1; }
		else	{ a = conf.geo.posx + conf.geo.height - peak; c = 1; }
		XFillRectangle(draw.display, draw.rootWin, draw.gc, a, b, c, d);
	}
}

// Our worker thread

void* worker_func() {
	long pixel[6];
	gint16 freq_data[2][256];
	double scale, x00, y00;
	unsigned int i, j, level;
	unsigned short *level1, *level2, *levelsw, *peak1, *peak2, *peakstep;
	
	level1 = (unsigned short*)calloc(256, sizeof(short)); // need to be zeroed out
	level2 = (unsigned short*)malloc(256*sizeof(short));
	peak1 = (unsigned short*)calloc(256, sizeof(short)); // need to be zeroed out
	peak2 = (unsigned short*)malloc(256*sizeof(short));
	peakstep = (unsigned short*)calloc(256, sizeof(short)); // need to be zeroed out
	if ((level1 == NULL)||(level2 == NULL)||(peak1 == NULL)||(peak2 == NULL)||(peakstep == NULL))
		error_exit("Allocation of memory failed");
	
	initialize_X();
	
	while (thread_control != STOP) {
		if (thread_control == CONF) {
			config_read();
			pixel[0] = get_color(conf.bar.color[0]);
			pixel[1] = get_color(conf.bar.color[1]);
			pixel[2] = get_color(conf.bar.color[2]);
			pixel[3] = get_color(conf.bar.color[3]);
			pixel[4] = get_color(conf.peak.color);
			pixel[5] = get_color(conf.bar.shadow_color);
			scale = conf.geo.height / (log((1 - conf.linearity) / conf.linearity) * 4);
			x00 = conf.linearity*conf.linearity*32768.0 / (2*conf.linearity - 1);
			y00 = -log(-x00) * scale;
			print_status("Ready to draw");
			thread_control = GO;
		}
		pthread_mutex_lock(&mutex1);
		memcpy(&freq_data, &shared_freq_data, sizeof(gint16)*2*256);
		pthread_mutex_unlock(&mutex1);

		/* instead of copying the old level array to the second array, 
		we just tell the first is now the second one */
		levelsw = level1;
		level1 = level2;
		level2 = levelsw;
		levelsw = peak1;
		peak1 = peak2;
		peak2 = levelsw;
	
		for (i = 0; i < conf.cutoff/conf.div; i++) {
			level = 0;
			for (j = i*conf.div; j < (i+1)*conf.div; j++)
				if (level < freq_data[0][j])
					level = freq_data[0][j];
			level = level * (i*conf.div + 1);
			level = ceil(log(level - x00)*scale + y00);
			if (level < conf.geo.height) {
				if ((level2[i] > conf.bar.falloff)&&(level < level2[i] - conf.bar.falloff))
					level1[i] = level2[i] - conf.bar.falloff;
				else	level1[i] = level;
			} else level1[i] = conf.geo.height;
			if (conf.peak.enabled) {
				if (level1[i] > peak2[i] - conf.peak.falloff) {
					peak1[i] = level1[i];
					peakstep[i] = 0;
				} else if (peakstep[i] == conf.peak.step)
					if (peak2[i] > conf.peak.falloff) peak1[i] = peak2[i] - conf.peak.falloff;
					else peak1[i] = 0;
				else {
					peak1[i] = peak2[i];
					peakstep[i]++;
				}
			}
			draw_bar(i, pixel[(int)ceil((float)level1[i] / conf.geo.height * 4) - 1], pixel[4], pixel[5], level1[i], level2[i], peak1[i], peak2[i]);
		}
		XFlush(draw.display);
		xmms_usleep(1000000 / conf.fps);
	}
	print_status("Worker thread: Exiting");
	XClearWindow(draw.display, draw.rootWin);
	XCloseDisplay(draw.display); 
	free(level1);
	free(level2);
	return NULL;
}


// da xmms functions

static void rootvis_init(void) {
	int rc1;
	print_status("Initializing");
	
	clean_data();		
	thread_control = CONF;
	if ((rc1 = pthread_create(&work_thread, NULL, worker_func, NULL))) {
		fprintf(stderr, "Thread creation failed: %d\n", rc1);
		error_exit("Thread creation failed");
	}
}

static void rootvis_cleanup(void) {
	print_status("Cleanup... ");
	thread_control = STOP;
	pthread_join(work_thread, NULL);
	print_status("Clean Exit");
}

static void rootvis_about(void)
{
	print_status("About");
}

static void rootvis_configure(void)
{
	print_status("Configuration trigger");
	thread_control = CONF;
}

static void rootvis_playback_start(void)
{
	print_status("Playback starting");
}

static void rootvis_playback_stop(void)
{
	clean_data();
}

static void rootvis_render_freq(gint16 freq_data[2][256]) {
	int channel, bucket;

	pthread_mutex_lock(&mutex1);
	for (channel = 0; channel < 2; channel++) {
	 for (bucket = 0; bucket < 256; bucket++) {
		shared_freq_data[channel][bucket] = freq_data[channel][bucket];
	 }
	}
	pthread_mutex_unlock(&mutex1);
}
